"""
eobox.raster subpackage

Subpackage which mainly builds upon rasterio (and thus gdal) functionality.
It provides tools for 

* extracting values of raster sampledata at location given by a vector dataset,
* chunk-wise processing (e.g. classification) of multiple single layer files.


For more information on the package content, visit [readthedocs](https://eo-box.readthedocs.raster/en/latest/eobox.raster.html).

"""

from .extraction import extract
from .extraction import load_extracted
from .rasterprocessing import MultiRasterIO
from .rasterprocessing import windows_from_blocksize



import numpy as np


dtype_ranges = {
    'uint8': (0, 255),
    'uint16': (0, 65535),
    'int16': (-32768, 32767),
    'uint32': (0, 4294967295),
    'int32': (-2147483648, 2147483647),
    'float32': (-3.4028235e+38, 3.4028235e+38),
    'float64': (-1.7976931348623157e+308, 1.7976931348623157e+308)}

def dtype_checker_df(df, dtype, return_=None):
    """Check if there are NaN values of values outside of a given datatype range.

    Arguments:
        df {dataframe} -- A dataframe.
        dtype {str} -- The datatype to check for.
    Keyword Arguments:
        return_ {str} -- Returns a boolean dataframe with the values not in the range of the dtype ('all'),
            the row ('rowsums') or column ('colsums') sums of that dataframe or an exit code 1 (None, default)
            if any of the values is not in the range.

    Returns:
        [int or DataFrame or Series] -- If no value is out of the range exit code 0 is returned, else depends on return_.
    """
    dtype_range = dtype_ranges[dtype]
    df_out_of_range = (df < dtype_range[0]) | (df > dtype_range[1]) | (~np.isfinite(df))
    if df_out_of_range.any().any():
        if return_== "colsums":
            df_out_of_range = df_out_of_range.apply(sum, axis=0) # column
        elif return_== "rowsums":
            df_out_of_range = df_out_of_range.apply(sum, axis=1) # row
        elif return_== "all":
            df_out_of_range = df_out_of_range
        else:
            df_out_of_range = 1
    else:
        df_out_of_range = 0
    return df_out_of_range

def cleanup_df_values_for_given_dtype(df, dtype, lower_as=None, higher_as=None, nan_as=None):
    def is_number(s):
        try:
            float(s)
            return True
        except ValueError:
            return False

    if dtype not in dtype_ranges.keys():
        raise ValueError("'dtype' must be one of {dtype_ranges.keys()}")

    dtype_range = dtype_ranges[dtype]
    lut = {"lower_as": lower_as,
           "higher_as": higher_as,
           "nan_as": nan_as,
           }
    for key, value in lut.items():
        # the default is to assign the maximum number for unsigned dtypes and the min value for signed dtypes
        # this is a common convention with raster data
        # the real values will be derived below
        if value is None:
            if dtype[0] == "u":
                value = "max_range"
            else:
                value = "min_range"

        # derive or validate the values to be filled in
        if isinstance(value, str):
            if value not in ["min_range", "max_range"]:
                raise ValueError(f"'*_as' parameters must be numeric or in case of a string 'min_range' or 'max'range'.")
            else:
                if value == "min_range":
                    lut[key] = dtype_range[0]
                else:
                    lut[key] = dtype_range[1]
        elif is_number(value):
            if (value < dtype_range[0]) | (value > dtype_range[1]):
                raise ValueError("Numeric value of the '*_as' parameters must be in the data range of the given dtype.")
            else:
                lut[key] = value
        else:
            raise ValueError(f"'*_as' parameters must be numeric or in case of a string 'min_range' or 'max'range'.")

    # print(f"Replacing mapping: {lut}")
    df = df.where(~(df < dtype_range[0]), lut["lower_as"])
    df = df.where(~(df > dtype_range[1]), lut["higher_as"])
    df = df.where(~(df.isna()), lut["nan_as"])
    df = df.astype(dtype)

    return df



"""
eobox.sampledata subpackage

The subpackage which holds data and functions to get file paths to data. 
sampledata is used in the examples and tests of other eobox sub-packages. 

For more information on the package content, visit [readthedocs](https://eo-box.readthedocs.raster/en/latest/eobox.sampledata.html).
"""

from . import base

from .base import get_dataset



"""
Module for extracting values of raster sampledata at location given by a vector dataset.
"""

import os
from pathlib import Path
import numpy as np
import pandas as pd
import rasterio
from tqdm import tqdm

from ..raster.gdalutils import rasterize

def extract(src_vector: str,
            burn_attribute: str,
            src_raster: list,
            dst_names: list,
            dst_dir: str,
            src_raster_template: str = None,
            gdal_dtype: int = 4,
            n_jobs: int = 1):
    """Extract values from list of single band raster for pixels overlapping with a vector data.

    The extracted data will be stored in the ``dst_dir`` by using the ``dst_names`` for the
    filename. If a file with a given name already exists the raster will be skipped.

    Arguments:
        src_vector {str} -- Filename of the vector dataset. Currently it must have the same CRS as
            the raster.
        burn_attribute {str} -- Name of the attribute column in the ``src_vector`` dataset to be
            stored with the extracted data. This should usually be a unique ID for the features
            (points, lines, polygons) in the vector dataset.
        src_raster {list} -- List of filenames of the single band raster files from which to
            extract.
        dst_names {list} -- List corresponding to ``src_raster`` names used to store and later
            identify the extracted to.
        dst_dir {str} -- Directory to store the data to.

    Keyword Arguments:
        src_raster_template {str} -- A template raster to be used for rasterizing the vectorfile.
            Usually the first element of ``src_raster``. (default: {None})
        gdal_dtype {int} -- Numeric GDAL data type, defaults to 4 which is UInt32.
            See https://github.com/mapbox/rasterio/blob/master/rasterio/dtypes.py for useful look-up
            tables.

    Returns:
        [int] -- If successful, 0 is returned as exit code.
    """
    if src_raster_template is None:
        src_raster_template = src_raster[0]
    path_rasterized = os.path.join(dst_dir, f"burn_attribute_rasterized_{burn_attribute}.tif")
    paths_extracted_aux = {ele: os.path.join(dst_dir, f"{ele}.npy") \
                           for ele in [f"aux_vector_{burn_attribute}",
                                       "aux_coord_x",
                                       "aux_coord_y"]}
    paths_extracted_raster = {}
    for path, name in zip(src_raster, dst_names):
        dst = f"{os.path.join(dst_dir, name)}.npy"
        if not os.path.exists(dst):
            paths_extracted_raster[path] = dst

    if not os.path.exists(dst_dir):
        os.makedirs(dst_dir)
    # if it does not already exist, here we first create the rasterized data
    if not os.path.exists(path_rasterized):
        if src_raster_template is None:
            src_raster_template = src_raster[0]
        # print("Rasterizing vector attribute.")
        rasterize(src_vector=src_vector,
                  burn_attribute=burn_attribute,
                  src_raster_template=src_raster_template,
                  dst_rasterized=path_rasterized,
                  gdal_dtype=gdal_dtype)

    # if any of the destination files do not exist we need the locations of the pixels to be
    # extracted in form of a numpy array bool (mask_arr) that fits the rasters from which we will
    # extract below
    if not (all([os.path.exists(path) for path in paths_extracted_aux.values()]) and \
            all([os.path.exists(path) for path in paths_extracted_raster.values()])):
        # print("Creating mask array for pixels to be extracted.")
        mask_arr = _get_mask_array(path_rasterized, paths_extracted_aux, burn_attribute)
    else:
        return 0

    # create the pixel coordinates if they do not exist
    if not all([os.path.exists(paths_extracted_aux["aux_coord_x"]),
                os.path.exists(paths_extracted_aux["aux_coord_y"])]):
        _create_and_save_coords(path_rasterized, paths_extracted_aux, mask_arr)

    # lets extract the raster values in case of sequential processing
    # or remove existing raster layers to prepare parallel processing
    if n_jobs == 1:
        for path_src, path_dst in tqdm(paths_extracted_raster.items(),
                                       total=len(paths_extracted_raster)):
            _extract_and_save_one_layer(path_src, path_dst, mask_arr)
    else:
        import multiprocessing as mp
        if n_jobs == -1:
            n_jobs = mp.cpu_count()
        pool = mp.Pool(processes=n_jobs)
        _ = [pool.apply_async(_extract_and_save_one_layer,
                              args=(src, dst, mask_arr)) for \
                                  src, dst in paths_extracted_raster.items()]
        pool.close()
        pool.join()
    return 0

def _get_mask_array(path_rasterized, paths_extracted_aux, burn_attribute):
    with rasterio.open(path_rasterized) as src:
        fids_arr = src.read()
        mask_arr = fids_arr > 0
        if not os.path.exists(paths_extracted_aux[f"aux_vector_{burn_attribute}"]):
            fids = fids_arr[mask_arr]
            del fids_arr
            np.save(paths_extracted_aux[f"aux_vector_{burn_attribute}"], fids)
            del fids
    return mask_arr

def _create_and_save_coords(path_rasterized, paths_extracted_aux, mask_arr):
    src = rasterio.open(path_rasterized)
    coords = {"x": rasterio.transform.xy(src.meta["transform"],
                                         rows=[0] * src.meta["width"],
                                         cols=np.arange(src.meta["width"]),
                                         offset='center')[0],
              "y": rasterio.transform.xy(src.meta["transform"],
                                         rows=np.arange(src.meta["height"]),
                                         cols=[0] * src.meta["height"],
                                         offset='center')[1]}
    coords_2d_array_x, coords_2d_array_y = np.meshgrid(coords["x"], coords["y"])
    del coords
    np.save(paths_extracted_aux["aux_coord_x"],
            np.expand_dims(coords_2d_array_x, axis=0)[mask_arr])
    del coords_2d_array_x
    np.save(paths_extracted_aux["aux_coord_y"],
            np.expand_dims(coords_2d_array_y, axis=0)[mask_arr])
    del coords_2d_array_y


def _extract_and_save_one_layer(path_src, path_dst, mask_arr):
    with rasterio.open(path_src) as src:
        raster_vals = src.read()[mask_arr]
        np.save(path_dst, raster_vals)


def load_extracted(src_dir: str,
                   patterns="*.npy",
                   vars_in_cols: bool = True,
                   index: pd.Series = None):
    """Load data extracted and stored by :py:func:`extract`

    Arguments:
        src_dir {str} -- The directory where the data is stored.

    Keyword Arguments:
        patterns {str, or list of str} -- A pattern (str) or list of patterns (list)
            to identify the variables to be loaded.
            The default loads all variables, i.e. all .npy files. (default: {'*.npy'})
        vars_in_cols {bool} -- Return the variables in columns (``True``) or rows ``False``
            (default: {True})
        index {pd.Series} -- A boolean pandas Series which indicates with ``True`` which samples to
            load.

    Returns:
        pandas.DataFrame -- A dataframe with the data.
    """
    def _load(path, index):
        if index is None:
            arr = np.load(str(path))
        else:
            arr = np.load(str(path), mmap_mode="r")[index]
        return arr

    src_dir = Path(src_dir)
    paths = []
    if isinstance(patterns, str):
        patterns = [patterns]
    for pat in patterns:
        paths += src_dir.glob(pat)

    if vars_in_cols:
        df_data = {}
        for path in paths:
            df_data[path.stem] = _load(path, index)
            df_data = pd.DataFrame(df_data)
            if index is not None:
                df_data.index = index.index[index]
    else:
        df_data = []
        for path in paths:
            arr = _load(path, index)
            df_data.append(pd.DataFrame(np.expand_dims(arr, 0), index=[path.stem]))
        df_data = pd.concat(df_data)
        if index is not None:
            df_data.columns = index.index[index]
    return df_data



import numpy as np
from pathlib import Path
import rasterio
from rasterio.warp import reproject
from rasterio.enums import Resampling
from rasterio import transform

class MultiRasterIO():
    def __init__(self, layer_files: list,
                 dst_res: float = None,
                 downsampler: str = 'average',
                 upsampler: str = 'nearest'):
        """Read, process and write multiple single layer raster files.

        Arguments:
            layer_files {list} -- List of paths to single band raster files.
            dst_res {float} -- Destination resolution. If `None` it is set to the highest resolution of all layers.

        Keyword Arguments:
            dst_res {numeric} -- Destination resolution in which to return the array.
                If `None`, each layer is returned in its native resolution,
                else each layer is resampled to the destination resolution `dst_res` (default: {None}).
            downsampler {str} -- The method used for downsampling layers with higher
                raster cell size (default: {'average'}).
            upsampler {str} -- The method used for upsampling layers with lower
                raster cell size  (default: {'nearest'})
        """

        # build the multi-resolution data frame here
        self._layer_files = layer_files
        self._layer_resolution = []  # make these public but with tupels ???
        self._layer_meta = []
        self._res_indices = {}
        self.windows = None
        self.windows_row = None
        self.windows_col = None
        for i, filename in enumerate(self._layer_files):
            with rasterio.open(filename) as src:
                res = src.meta["transform"][0]
                self._layer_resolution.append(res)
                self._layer_meta.append(src.meta)
            if res not in self._res_indices.keys():
                self._res_indices[res] = [i]
            else:
                self._res_indices[res].append(i)
        self.dst_res = self._get_dst_resolution(dst_res)
        self._windows_res = self.dst_res
        self.downsampler = downsampler
        self.upsampler = upsampler

    # def windows(self, windows, resolution): # setter and getter ?
    #     raise NotImplementedError()
        # ...
        # self.windows = ...

    def _get_dst_resolution(self, dst_res=None):
        """Get default resolution, i.e. the highest resolution or smallest cell size."""
        if dst_res is None:
            dst_res = min(self._res_indices.keys())
        return dst_res

    def block_windows(self, res=None):  # setter and getter ?
        """Load windows for chunks-wise processing from raster internal tiling (first raster of given resolution).

        Arguments:
            res {numeric} -- Resolution determining the raster (1st of resolution group) from which to take the tiling.
        """

        if res is None:
            res = max(self._res_indices.keys())
        self._windows_res = res
        a_file_index_given_res = self._res_indices[res][0]
        with rasterio.open(self._layer_files[a_file_index_given_res]) as src:
            wins_of_first_dst_res_layer = tuple(src.block_windows())
        self.windows = np.array([win[1] for win in wins_of_first_dst_res_layer])
        self.windows_row = np.array([win[0][0] for win in wins_of_first_dst_res_layer])
        self.windows_col = np.array([win[0][1] for win in wins_of_first_dst_res_layer])

    def windows_from_blocksize(self, blocksize_xy=512):
        """Create rasterio.windows.Window instances with given size which fully cover the raster.

        Arguments:
            blocksize_xy {int or list of two int} -- Size of the window. If one integer is given it defines
                the width and height of the window. If a list of two integers if given the first defines the
                width and the second the height.

        Returns:
            None -- But the attributes ``windows``, ``windows_row`` and ``windows_col`` are updated.
        """
        meta = self._get_template_for_given_resolution(self.dst_res, "meta")
        width = meta["width"]
        height = meta["height"]
        blocksize_wins = windows_from_blocksize(blocksize_xy, width, height)

        self.windows = np.array([win[1] for win in blocksize_wins])
        self.windows_row = np.array([win[0][0] for win in blocksize_wins])
        self.windows_col = np.array([win[0][1] for win in blocksize_wins])
        return self

    def _get_template_for_given_resolution(self, res, return_):
        """Given specified resolution ('res') return template layer 'path', 'meta' or 'windows'."""
        path = self._layer_files[self._res_indices[res][0]]
        if return_ == "path":
            return_value = path
        else:
            with rasterio.open(str(path)) as src:
                if return_ == "meta":
                    return_value = src.meta
                elif return_ == "windows":
                    return_value = tuple(src.block_windows())
                else:
                    raise ValueError("'return_' must be 'path', meta' or 'windows'.")
        return return_value

    def windows_df(self):
        """Get Windows (W) W-row, W-col and W-index of windows e.g. loaded with :meth:`block_windows` as a dataframe.

        Returns:
            [dataframe] -- A dataframe with the window information and indices (row, col, index).
        """

        import pandas as pd
        if self.windows is None:
            raise Exception("You need to call the block_windows or windows before.")
        df_wins = []
        for row, col, win in zip(self.windows_row, self.windows_col, self.windows):
            df_wins.append(pd.DataFrame({"row":[row], "col":[col], "Window":[win]}))
        df_wins = pd.concat(df_wins).set_index(["row", "col"])
        df_wins["window_index"] = range(df_wins.shape[0])
        df_wins = df_wins.sort_index()
        return df_wins

    def ji_windows(self, ij_win):  # what can be given to ij_win NOT intuitive/right name by now!!!
        """For a given specific window, i.e. an element of :attr:`windows`, get the windows of all resolutions.

        Arguments:
            ij_win {int} -- The index specifying the window for which to return the resolution-windows.
        """
        ji_windows = {}
        transform_src = self._layer_meta[self._res_indices[self._windows_res][0]]["transform"]
        for res in self._res_indices:
            transform_dst = self._layer_meta[self._res_indices[res][0]]["transform"]
            ji_windows[res] = window_from_window(window_src=self.windows[ij_win],
                                                 transform_src=transform_src,
                                                 transform_dst=transform_dst)
        return ji_windows

    def get_arrays(self, ji_win):
        """Get the data of the a window given the ji_windows derived with :method:`ji_windows`.

        Arguments:
            ji_win {[type]} -- The index of the window or the (multi-resolution) windows returned by :meth:`ji_window`.

        Returns:
            (list of) array(s) -- List of 2D arrays in native resolution in case `dst_res` is `None`
                or a 3D array where all layers are resampled to `dst_res` resolution.
        """
        if isinstance(ji_win, dict):
            ji_windows = ji_win
        else:
            ji_windows = self.ji_windows(ji_win)

        arrays = []
        for filename, res in zip(self._layer_files, self._layer_resolution):
            with rasterio.open(filename) as src:
                arr = src.read(1, window=ji_windows[res])
            arrays.append(arr)
        if self.dst_res is not None:
            arrays = self._resample(arrays=arrays, ji_windows=ji_windows)
        return arrays

    def _resample(self, arrays, ji_windows):
        """Resample all arrays with potentially different resolutions to a common resolution."""
        # get a destination array template
        win_dst = ji_windows[self.dst_res]
        aff_dst = self._layer_meta[self._res_indices[self.dst_res][0]]["transform"]
        arrays_dst = list()
        for i, array in enumerate(arrays):
            arr_dst = np.zeros((int(win_dst.height), int(win_dst.width)))
            if self._layer_resolution[i] > self.dst_res:
                resampling = getattr(Resampling, self.upsampler)
            elif self._layer_resolution[i] < self.dst_res:
                resampling = getattr(Resampling, self.downsampler)
            else:
                arrays_dst.append(array.copy())
                continue
            reproject(array, arr_dst,  # arr_dst[0, :, :, i],
                      src_transform=self._layer_meta[i]["transform"],
                      dst_transform=aff_dst,
                      src_crs=self._layer_meta[0]["crs"],
                      dst_crs=self._layer_meta[0]["crs"],
                      resampling=resampling)
            arrays_dst.append(arr_dst.copy())
        arrays_dst = np.stack(arrays_dst, axis=2) # n_images x n x m x 10 would be the synergise format
        return arrays_dst

    def apply_and_save(self, dst_files, func, **kwargs):

        result_0 = self._process_window(0, func, **kwargs)
        if len(dst_files) != len(result_0):
            raise ValueError("The number of file paths in 'dst' need to match the number of output layers.")
        meta = self._get_template_for_given_resolution(res=self.dst_res, return_="meta")

        dtypes = [arr.dtype for arr in result_0]
        dsts = []
        for layer_idx, dst_name_dtype in enumerate(zip(dst_files, dtypes)):
            dst_name, dtype = dst_name_dtype
            Path(dst_name).parent.mkdir(parents=True, exist_ok=True)
            meta.update(dtype=dtype)
            dst = rasterio.open(str(dst_name), "w", **meta)
            ji_windows = self.ji_windows(0)
            dst.write(result_0[layer_idx], 1, window=ji_windows[self.dst_res])
            dsts.append(dst)

        for ji_win in range(1, len(self.windows)):
            result_ji = self._process_window(ji_win, func, **kwargs)
            for layer_idx, dtype in enumerate(dtypes):
                Path(dst_name).parent.mkdir(parents=True, exist_ok=True)
                meta.update(dtype=dtype)
                ji_windows = self.ji_windows(ji_win)
                dsts[layer_idx].write(result_ji[layer_idx], 1, window=ji_windows[self.dst_res])

        for dst in dsts:
            dst.close()

        return 0

    def _process_windows_merge_stack(self, func, **kwargs):
        """Load (resampled) array of all windows, apply custom function on it, merge and stack results to one array."""
        ji_results = self._process_windows(func, **kwargs)
        for idx_layer in range(len(ji_results[0])):  # this is the number of output layers
            for j in np.unique(self.windows_row):
                win_indices_j = np.where(self.windows_row == j)[0]
                layer_merged_j = np.hstack([ji_results[idx][idx_layer] for idx in win_indices_j])
                if j == 0:
                    layer_merged = layer_merged_j
                else:
                    layer_merged = np.vstack([layer_merged, layer_merged_j])
            if idx_layer == 0:
                layers_merged = layer_merged
            else:
                layers_merged = np.stack([layers_merged, layer_merged], axis=2)
        return layers_merged

    def _process_windows(self, func, **kwargs):
        """Load (resampled) array of all windows and apply custom function on it."""
        ji_results = []
        for ji_win in range(len(self.windows)):
            ji_results.append(self._process_window(ji_win, func, **kwargs))
        return ji_results

    def _process_window(self, ji_win, func, **kwargs):
        """Load (resampled) array of window ji_win and apply custom function on it. """
        arr = self.get_arrays(ji_win)
        result = func(arr, **kwargs)
        return result

    def get_window_from_xy(self, xy):
        """Get the window index given a coordinate (raster CRS)."""
        a_transform = self._get_template_for_given_resolution(res=self.dst_res, return_="meta")["transform"]
        row, col = transform.rowcol(a_transform, xy[0], xy[1])
        ij_containing_xy = None
        for ji, win in enumerate(self.windows):
            (row_start, row_end), (col_start, col_end) = rasterio.windows.toranges(win)
            # print(row, col, row_start, row_end, col_start, col_end)
            if ((col >= col_start) & (col < col_end)) & ((row >= row_start) & (row < row_end)):
                ij_containing_xy = ji
                break
        if ij_containing_xy is None:
            raise ValueError("The given 'xy' value is not contained in any window.")
        return ij_containing_xy

def window_from_window(window_src, transform_src, transform_dst):
    # extend that transform can be a filename, rasterio dataset, meta (dict), or Affine
    spatial_bounds = rasterio.windows.bounds(window=window_src, transform=transform_src,
                                             height=0, width=0)  # defaults
    window_dst = rasterio.windows.from_bounds(spatial_bounds[0],
                                              spatial_bounds[1],
                                              spatial_bounds[2],
                                              spatial_bounds[3],
                                              transform=transform_dst,
                                              height=None, width=None, precision=None)  # defaults
    return window_dst

def windows_from_blocksize(blocksize_xy, width, height):
    """Create rasterio.windows.Window instances with given size which fully cover a raster.

    Arguments:
        blocksize_xy {int or list of two int} -- [description]
        width {int} -- With of the raster for which to create the windows.
        height {int} -- Heigth of the raster for which to create the windows.

    Returns:
        list -- List of windows according to the following format
            ``[[<row-index>, <column index>], rasterio.windows.Window(<col_off>, <row_off>, <width>, <height>)]``.
    """

    # checks the blocksize input
    value_error_msg = "'blocksize must be an integer or a list of two integers.'"
    if isinstance(blocksize_xy, int):
        blockxsize, blockysize = (blocksize_xy, blocksize_xy)
    elif isinstance(blocksize_xy, list):
        if len(blocksize_xy) != 2:
            raise ValueError(value_error_msg)
        else:
            if not all([isinstance(blocksize_xy[0], int), isinstance(blocksize_xy[1], int)]):
                raise ValueError(value_error_msg)
            blockxsize, blockysize = blocksize_xy
    else:
        raise ValueError(value_error_msg)

    # create the col_off and row_off elements for all windows
    n_cols = int(np.ceil(width / blockxsize))
    n_rows = int(np.ceil(height / blockysize))
    col = list(range(n_cols)) * n_rows
    col_off = np.array(col) * blockxsize
    row = np.repeat(list(range(n_rows)), n_cols)
    row_off = row * blockysize

    # create the windows
    # if necessary, reduce the width and/or height of the border windows
    blocksize_wins = []
    for ridx, roff, cidx, coff, in zip(row, row_off, col, col_off):
        if coff + blockxsize > width:
            bxsize = width - coff
        else:
            bxsize = blockxsize
        if roff + blockysize > height:
            bysize = height - roff
        else:
            bysize = blockysize
        blocksize_wins.append([[ridx, cidx], rasterio.windows.Window(coff, roff, bxsize, bysize)])
    return blocksize_wins



from os.path import relpath
from osgeo import gdal, gdalconst, ogr
from pathlib import Path
import warnings

# find the path to the gdal_proximity.py - I found it in these places so far...
try:
    PROXIMITY_PATH = str(list(Path(gdal.__file__).parent.rglob("*proximity.py"))[0])
except:
    PROXIMITY_PATH = None
try:
    PROXIMITY_PATH = str(list(Path(gdal.__file__).parent.parent.rglob("GDAL*/scripts/*proximity.py"))[0])
except:
    PROXIMITY_PATH = None

if PROXIMITY_PATH is None:
    warnings.warn("Could not find the path of gdal_proximity.py: Searched in " + \
                    f"{Path(gdal.__file__).parent}, {str(Path(gdal.__file__).parent.parent)+'/GDAL*/scripts'}.")


def buildvrt(input_file_list, output_file,
             relative=True, **kwargs):
    """Build a VRT

    See also: https://www.gdal.org/gdalbuildvrt.html

    You can find the possible BuildVRTOptions (**kwargs**) here:
    https://github.com/nextgis/pygdal/blob/78a793057d2162c292af4f6b240e19da5d5e52e2/2.1.0/osgeo/gdal.py#L1051

    Arguments:
        input_file_list {list of str or Path objects} -- List of input files.
        output_file {str or Path object} -- Output file (VRT).

    Keyword Arguments:
        relative {bool} -- If ``True``, the ``input_file_list`` paths are converted to relative
            paths (relative to the output file) and the VRT works even if the data is moved somewhere else -
            given that the relative location of theVRT and the input files does not chance!
        **kwargs {} -- BuildVRTOptions - see function description for a link to .

    Returns:
        [int] -- If successful, 0 is returned as exit code.
    """

    # create destination directory
    if not Path(output_file).parent.exists():
        Path(output_file).parent.mkdir(parents=True, exist_ok=True)

    # make sure we have absolute paths and strings since BuildVRT does not like something else
    input_file_list = [str(Path(p).absolute()) for p in input_file_list]
    output_file = str(Path(output_file).absolute())

    vrt_options = gdal.BuildVRTOptions(**kwargs)

    vrt = gdal.BuildVRT(output_file,
                        input_file_list,
                        options=vrt_options)
    vrt = None

    # if needed, create the input file paths relative to the output vrt path
    # and replace them in the vrt.
    # if desired, fix the paths and the relativeToVRT tag in the VRT
    if relative:
        input_file_list_relative = [relpath(p, Path(output_file).parent) for p in input_file_list]

        with open(output_file, 'r') as file:
            # read a list of lines into data
            lines = file.readlines()

        new_lines = []
        counter = -1
        for line in lines:
            # sometimes it is relative by default
            # maybe when all files contain the parent directory of the output file (?)
            if "relativeToVRT=\"1\"" in line:
                counter += 1
            elif "relativeToVRT=\"0\"" in line:
                counter += 1
                input_file = str(input_file_list[counter])
                input_file_relative = str(input_file_list_relative[counter])
                if input_file not in line:
                    raise Exception(f"Expect path {input_file} not part of line {line}.")
                line = line.replace(input_file,
                                    input_file_relative)
                line = line.replace("relativeToVRT=\"0\"",
                                    "relativeToVRT=\"1\"")
            else:
                pass
            new_lines.append(line)

        with open(output_file, 'w') as file:
            file.writelines(new_lines)
    return 0

def reproject_on_template_raster(src_file, dst_file, template_file, resampling="near", compress=None, overwrite=False):
    """Reproject a one-band raster to fit the projection, extend, pixel size etc. of a template raster.  
    
    Function based on https://stackoverflow.com/questions/10454316/how-to-project-and-resample-a-grid-to-match-another-grid-with-gdal-python

    Arguments:
        src_file {str} -- Filename of the source one-band raster. 
        dst_file {str} -- Filename of the destination raster. 
        template_file {str} -- Filename of the template raster.
        resampling {str} -- Resampling type:
             'near' (default), 'bilinear', 'cubic', 'cubicspline', 'lanczos', 'average', 'mode', 'max', 'min', 'med', 'q1', 'q3',
            see https://www.gdal.org/gdalwarp.html -r parameter.
        compress {str} -- Compression type: None (default), 'lzw', 'packbits', 'defalte'.
    """
    
    if not overwrite and Path(dst_file).exists():
        print("Processing skipped. Destination file exists.")
        return 0
    
    GDAL_RESAMPLING_ALGORITHMS = {
        "bilinear": "GRA_Bilinear",
        "cubic": "GRA_Cubic",
        "cubicspline": "GRA_CubicSpline",
        "lanczos": "GRA_Lanczos",
        "average": "GRA_Average",
        "mode": "GRA_Mode",
        "max": "GRA_Max",
        "min": "GRA_Min",
        "med": "GRA_Med",
        "near": "GRA_NearestNeighbour",
        "q1": "GRA_Q1",
        "q3": "GRA_Q3"
    }

    compressions = ["lzw", "packbits", "deflate"]

    if resampling not in GDAL_RESAMPLING_ALGORITHMS.keys():
        raise ValueError(f"'resampling must be one of {', '.join(GDAL_RESAMPLING_ALGORITHMS.keys())}")

    if compress is None:
        options = []
    else:
        if compress.lower() not in compressions:
            raise ValueError(f"'compress must be one of {', '.join(compressions)}")
        else:
            options = [f'COMPRESS={compress.upper()}']
    
    # Source
    src = gdal.Open(src_file, gdalconst.GA_ReadOnly)
    src_band = src.GetRasterBand(1)
    src_proj = src.GetProjection()

    # We want a section of source that matches this:
    match_ds = gdal.Open(template_file, gdalconst.GA_ReadOnly)
    match_proj = match_ds.GetProjection()
    match_geotrans = match_ds.GetGeoTransform()
    wide = match_ds.RasterXSize
    high = match_ds.RasterYSize

    # Output / destination
    Path(dst_file).parent.mkdir(parents=True, exist_ok=True)
    dst = gdal.GetDriverByName('GTiff').Create(dst_file, wide, high, 1, src_band.DataType, options=options)
    dst.SetGeoTransform( match_geotrans )
    dst.SetProjection( match_proj)

    # Do the work
    gdal.ReprojectImage(src, dst, src_proj, match_proj, 
                        getattr(gdalconst, GDAL_RESAMPLING_ALGORITHMS[resampling]))


    del dst # Flush
    return 0

def rasterize(src_vector: str,
              burn_attribute: str,
              src_raster_template: str,
              dst_rasterized: str,
              gdal_dtype: int = 4):
    """Rasterize the values of a spatial vector file.

    Arguments:
        src_vector {str}} -- A OGR vector file (e.g. GeoPackage, ESRI Shapefile) path containing the
            data to be rasterized.
        burn_attribute {str} -- The attribute of the vector data to be burned in the raster.
        src_raster_template {str} -- Path to a GDAL raster file to be used as template for the
            rasterized data.
        dst_rasterized {str} -- Path of the destination file.
        gdal_dtype {int} -- Numeric GDAL data type, defaults to 4 which is UInt32.
            See https://github.com/mapbox/rasterio/blob/master/rasterio/dtypes.py for useful look-up
            tables.
    Returns:
        None
    """

    data = gdal.Open(str(src_raster_template),  # str for the case that a Path instance arrives here
                     gdalconst.GA_ReadOnly)
    geo_transform = data.GetGeoTransform()
    #source_layer = data.GetLayer()
    # x_max = x_min + geo_transform[1] * data.RasterXSize
    # y_min = y_max + geo_transform[5] * data.RasterYSize
    x_res = data.RasterXSize
    y_res = data.RasterYSize
    mb_v = ogr.Open(src_vector)
    mb_l = mb_v.GetLayer()
    target_ds = gdal.GetDriverByName('GTiff').Create(dst_rasterized,
                                                     x_res, y_res, 1,
                                                     gdal_dtype)  # gdal.GDT_Byte
    # import osr
    target_ds.SetGeoTransform((geo_transform[0],  # x_min
                               geo_transform[1],  # pixel_width
                               0,
                               geo_transform[3],  # y_max
                               0,
                               geo_transform[5]  # pixel_height
                               ))
    prj = data.GetProjection()
    # srs = osr.SpatialReference(wkt=prj)  # Where was this needed?
    target_ds.SetProjection(prj)
    band = target_ds.GetRasterBand(1)
    # NoData_value = 0
    # band.SetNoDataValue(NoData_value)
    band.FlushCache()
    gdal.RasterizeLayer(target_ds, [1], mb_l, options=[f"ATTRIBUTE={burn_attribute}"])

    target_ds = None



""""
Base module for loading sampledata.
"""

import glob
import os

DIR_DATA = os.path.join(
    os.path.dirname(
        os.path.abspath(__file__)), "data")

def get_dataset(dataset="s2l1c"):
    """Get a specific sampledata to play around.

    So far the following sampledata exist:

    * 's2l1c': One Sentinel-2 Level 1C scene with a reference dataset.
    * 'lsts': A time series of 105 Landsat scenes each with the bands b3 (red), b4 (nir), b5 (swir1) and fmask.

    Keyword Arguments:
        dataset {str} -- The name of the dataset (default: {'s2l1c'}).

    Returns:
        [dict] -- A dictionary with paths and information about the sampledata.
    """
    if dataset == "s2l1c":
        search_string = os.path.join(DIR_DATA, dataset, "**", "*_B??.jp2")
        files = glob.glob(search_string, recursive=True)
        if not files:
            raise IOError(f"Could not find raster files of the s2l1c dataset. Search string: {search_string}")
        basename_splitted = [pth.replace(".jp2", "").split("_")[-2:] for pth in files]
        dset = {"raster_files": files,
                "raster_bands": [ele[1] for ele in basename_splitted],
                "raster_times": [ele[0] for ele in basename_splitted],
                "vector_file": os.path.join(DIR_DATA, "s2l1c", "s2l1c_ref.gpkg"),
                "vector_file_osm": os.path.join(DIR_DATA, "s2l1c", "gis_osm_landuse-water_a_free_1_area-10000-to-500000.gpkg")}

    elif dataset == "lsts":
        search_string = os.path.join(DIR_DATA, dataset, "**", "*.tif")
        files = glob.glob(search_string, recursive=True)
        if not files:
            raise IOError(f"Could not find raster files of the lsts dataset. Search string: {search_string}")
        basename_splitted = [os.path.basename(pth).replace(".tif", "").split("_") for pth in files]
        dset = {"raster_files": files,
                "raster_bands": [ele[1] for ele in basename_splitted],
                "raster_times": [ele[0][9:16] for ele in basename_splitted]}

    # If you want to add a new dataset here, do not forget to do all of the following steps:
    # 1) add the dataset in the eo-box/sampledata/eobox/sampledata/data/<name of new dataset>
    # 2) write the code here to get the paths of the data and eventually some additional information
    # 3) write a test to make sure you get the data
    # 4) add the new dataset to package_data in eo-box/sampledata/eobox/setup.py
    # 5) add the new dataset to package_data in eo-box/sampledata/MANIFEST.in
    # 4) change the version number in eo-box/sampledata/eobox/sampledata/__init__.py to '<current>.<current+1>.0'

    return dset



import geopandas as gpd 
from pathlib import Path
from osgeo import gdal
import rasterio
import shutil
import subprocess
import tempfile
from tqdm import tqdm

from ..raster.gdalutils import rasterize
from ..raster.gdalutils import PROXIMITY_PATH


def calc_distance_to_border(polygons, template_raster, dst_raster, overwrite=False,
                            keep_interim_files=False):
    """Calculate the distance of each raster cell (in and outside the polygons) to the next polygon border.
    
    Arguments:
        polygons {str} -- Filename to a geopandas-readable file with polygon features. 
        template_raster {[type]} -- Filename to a rasterio-readable file.
        dst_raster {[type]} -- Destination filename for the distance to polygon border raster file (tif).
    
    Keyword Arguments:
        overwrite {bool} -- Overwrite files if they exists? (default: {False})
        keep_interim_files {bool} -- Keep the interim line vector and raster files (default: {True})
    
    Returns:
        [type] -- [description]
    """
    if Path(dst_raster).exists() and not overwrite:
        print(f"Returning 0 - File exists: {dst_raster}")
        return 0

    with rasterio.open(template_raster) as tmp:
        crs = tmp.crs
        
    dst_raster = Path(dst_raster)
    dst_raster.parent.mkdir(exist_ok=True, parents=True)

    tempdir = Path(tempfile.mkdtemp(prefix=f"TEMPDIR_{dst_raster.stem}_", dir=dst_raster.parent))
    interim_file_lines_vector = tempdir / "interim_sample_vector_dataset_lines.shp"
    interim_file_lines_raster = tempdir / "interim_sample_vector_dataset_lines.tif"

    exit_code = convert_polygons_to_lines(polygons, 
                                          interim_file_lines_vector, 
                                          crs=crs, 
                                          add_allone_col=True)
    
    rasterize(src_vector=str(interim_file_lines_vector),
              burn_attribute="ALLONE",
              src_raster_template=str(template_raster),
              dst_rasterized=str(interim_file_lines_raster),
              gdal_dtype=1)

    cmd = f"{PROXIMITY_PATH} " \
          f"{str(Path(interim_file_lines_raster).absolute())} " \
          f"{str(Path(dst_raster).absolute())} " \
          f"-ot Float32 -distunits PIXEL -values 1 -maxdist 255"
    subprocess.check_call(cmd, shell=True)
    
    if not keep_interim_files:
        shutil.rmtree(tempdir)
    else:
        print(f"Interim files are in {tempdir}")
    return 0

def convert_polygons_to_lines(src_polygons, dst_lines, crs=None, add_allone_col=False):
    """Convert polygons to lines.

    Arguments:
        src_polygons {path to geopandas-readable file} -- Filename of the the polygon vector dataset to be 
            converted to lines.
        dst_lines {[type]} -- Filename where to write the line vector dataset to.

    Keyword Arguments:
        crs {dict or str} -- Output projection parameters as string or in dictionary format.
            This will reproject the data when a crs is given (not {None}) (default: {None}).
        add_allone_col {bool} -- Add an additional attribute column with all ones.
            This is useful, e.g. in case you want to use the lines with gdal_proximity afterwards (default: {True}).

    Returns:
        int -- Exit code 0 if successeful.
    """
    gdf = gpd.read_file(src_polygons)
    geom_coords = gdf["geometry"] # featureset.get(5)["geometry"]["coordinates"]
    lines = []
    row_ids = []
    for i_row, pol in tqdm(enumerate(geom_coords), total=len(geom_coords)):
        boundary = pol.boundary
        if boundary.type == 'MultiLineString':
            for line in boundary:
                lines.append(line)
                row_ids.append(i_row)
        else:
            lines.append(boundary)
            row_ids.append(i_row)

    gdf_lines = gdf.drop("geometry", axis=1).iloc[row_ids, :]
    gdf_lines["Coordinates"] = lines
    gdf_lines = gpd.GeoDataFrame(gdf_lines, geometry='Coordinates', crs=gdf.crs)
    if crs is not None:
        gdf_lines = gdf_lines.to_crs(crs)
    if add_allone_col:
        gdf_lines["ALLONE"] = 1
    Path(dst_lines).parent.mkdir(exist_ok=True, parents=True)
    gdf_lines.to_file(dst_lines)
    return 0



import pandas as pd
from pathlib import Path
import rasterio
from tqdm import tqdm

from eobox.raster import MultiRasterIO
from eobox.raster.gdalutils import buildvrt


class EOCube:
    def __init__(self, df_layers, chunksize=2 ** 5, wdir=None):
        self._df_layers = df_layers
        self._chunksize = chunksize
        # In this whole class we could directly get the windows from the
        # eobox.raster.windows_from_blocksize(blocksize_xy, width, height)
        # then we do not need the more expensive initialization of MultiRasterIO
        self._mrio = None
        self._initialize_mrio_if_attr_is_none()
        if wdir:
            self._wdir = Path(wdir)
        else:
            self._wdir = wdir

    @property
    def df_layers(self):
        return self._df_layers

    @property
    def chunksize(self):
        return self._chunksize

    @chunksize.setter
    def chunksize(self, value):
        self._mrio = self._mrio.windows_from_blocksize(value)
        self._chunksize = value

    @property
    def n_chunks(self):
        return len(self._mrio.windows)

    @property
    def wdir(self):
        return self._wdir

    @wdir.setter
    def wdir(self, value):
        self._wdir = value

    def _initialize_mrio_if_attr_is_none(self):
        if not self._mrio:
            self._mrio = MultiRasterIO(layer_files=self.df_layers["path"].values)
            self._mrio = self._mrio.windows_from_blocksize(self.chunksize)

    def get_chunk(self, ji):
        """Get a EOCubeChunk"""
        return EOCubeChunk.from_eocube(self, ji)

    def apply_and_write(self, fun, dst_paths, **kwargs):
        ji_process = []
        for ji in range(self.n_chunks):
            dst_paths_ji_exist = [
                self.get_chunk_path_from_layer_path(dst, ji, mkdir=False).exists()
                for dst in dst_paths
            ]
            if not all(dst_paths_ji_exist):
                ji_process.append(ji)
        if len(ji_process) != self.n_chunks:
            print(f"{self.n_chunks - len(ji_process)} chunks already processed and skipped.")
        results = []
        for ji in tqdm(ji_process, total=len(ji_process)):
            eoc_chunk = self.get_chunk(ji)
            results.append(fun(eoc_chunk, dst_paths, **kwargs))
        for pth in dst_paths:
            if not Path(pth).exists():
                self.create_vrt_from_chunks(pth)

    def get_chunk_path_from_layer_path(self, path_layer, ji, mkdir=True):
        path_layer = Path(path_layer)
        bdir_chunks = path_layer.parent / f"xchunks_cs{self.chunksize}" / path_layer.stem
        if mkdir:
            bdir_chunks.mkdir(exist_ok=True, parents=True)
        ndigits = len(str((self.n_chunks)))
        dst_path_chunk = bdir_chunks / (path_layer.stem + f"_ji-{ji:0{ndigits}d}.tif")
        return dst_path_chunk

    def create_vrt_from_chunks(self, dst_path):
        chunk_paths = []
        for ji in range(self.n_chunks):
            pth = self.get_chunk_path_from_layer_path(dst_path, ji, mkdir=True).absolute()
            if not pth.exists():
                raise FileNotFoundError(pth)
            else:
                chunk_paths.append(str(pth))
        buildvrt(chunk_paths, str(dst_path), relative=True)


class EOCubeChunk(EOCube):
    def __init__(self, ji, df_layers, chunksize=2 ** 5, wdir=None):
        super().__init__(df_layers=df_layers, chunksize=chunksize, wdir=wdir)
        self._ji = ji
        self._data = None  # set with self.read_data()
        self._data_structure = None  # can be ndarray, dataframe
        self._window = self._mrio.windows[self.ji]
        self._width = self._window.width
        self._height = self._window.height
        self._n_layers = self.df_layers.shape[0]
        self.result = None

    @property
    def ji(self):
        return self._ji

    @property
    def data(self):
        return self._data

    @property
    def chunksize(self):
        return self._chunksize

    @chunksize.setter
    def chunksize(self, value):
        raise NotImplementedError(
            "It is not allowed to set the EOCubeChunk chunksize (but of a EOCube object)."
        )

    def read_data(self):
        self._data = self._mrio.get_arrays(self.ji)
        self._update_data_structure()
        return self

    def _update_data_structure(self):
        self._data_structure = self.data.__class__.__name__

    def convert_data_to_dataframe(self):
        if self._data_structure != "ndarray":
            raise Exception(f"Data is not an ndarray but {self._data_structure}.")
        if "uname" not in self.df_layers.columns:
            raise Exception("You need a column named uname with unique names for all layers.")
        if not self.df_layers["uname"].is_unique:
            raise Exception("You need a column named uname with unique names for all layers.")

        self._data = pd.DataFrame(self._reshape_3to2d(self._data))
        self._update_data_structure()
        self._data.columns = self.df_layers["uname"]
        return self

    def convert_data_to_ndarray(self):
        """Converts the data from dataframe to ndarray format. Assumption: df-columns are ndarray-layers (3rd dim.)"""
        if self._data_structure != "DataFrame":
            raise Exception(f"Data is not a DataFrame but {self._data_structure}.")
        self._data = self._convert_to_ndarray(self._data)
        self._update_data_structure()
        return self

    def _convert_to_ndarray(self, data):
        """Converts data from dataframe to ndarray format. Assumption: df-columns are ndarray-layers (3rd dim.)"""
        if data.__class__.__name__ != "DataFrame":
            raise Exception(f"data is not a DataFrame but {data.__class__.__name__}.")
        shape_ndarray = (self._height, self._width, data.shape[1])
        data_ndarray = data.values.reshape(shape_ndarray)
        return data_ndarray

    def write_dataframe(self, result, dst_paths, nodata=None, compress="lzw"):
        """Write results (dataframe) to disc."""
        result = self._convert_to_ndarray(result)
        self.write_ndarray(result, dst_paths, nodata=nodata, compress=compress)

    def write_ndarray(self, result, dst_paths, nodata=None, compress="lzw"):
        """Write results (ndarray) to disc."""

        assert len(dst_paths) == result.shape[2]
        assert result.shape[0] == self._height
        assert result.shape[1] == self._width
        assert result.shape[2] == len(dst_paths)
        with rasterio.open(
            self._mrio._get_template_for_given_resolution(self._mrio.dst_res, "path")
        ) as src_layer:
            pass  # later we need src_layer for src_layer.window_transform(win)
        for i, pth in enumerate(dst_paths):
            dst_path_chunk = self.get_chunk_path_from_layer_path(pth, self.ji)

            result_layer_i = result[:, :, [i]]
            assert result_layer_i.shape[2] == 1
            kwargs = self._mrio._get_template_for_given_resolution(
                res=self._mrio.dst_res, return_="meta"
            ).copy()
            kwargs.update(
                {
                    "driver": "GTiff",
                    "compress": compress,
                    "nodata": nodata,
                    "height": self._height,
                    "width": self._width,
                    "dtype": result_layer_i.dtype,
                    "transform": src_layer.window_transform(self._window),
                }
            )
            with rasterio.open(dst_path_chunk, "w", **kwargs) as dst:
                dst.write(result_layer_i[:, :, 0], 1)

    @staticmethod
    def _reshape_3to2d(arr_3d):
        new_shape = (arr_3d.shape[0] * arr_3d.shape[1], arr_3d.shape[2])
        return arr_3d.reshape(new_shape)

    @staticmethod
    def from_eocube(eocube, ji):
        """Create a EOCubeChunk object from an EOCube object."""
        eocubewin = EOCubeChunk(ji, eocube.df_layers, eocube.chunksize, eocube.wdir)
        return eocubewin



__version__ = "0.3.1"

from . import sampledata
from .sampledata import get_dataset

from . import raster
from .raster import extract
from .raster import load_extracted
from .raster import MultiRasterIO

from . import vector
from .vector import convert_polygons_to_lines
from .vector import calc_distance_to_border

__path__ = __import__('pkgutil').extend_path(__path__, __name__)



import numpy as np
import pandas as pd
from pathlib import Path
import rasterio
import time
from tqdm import tqdm

from .rasterprocessing import MultiRasterIO
from .gdalutils import buildvrt

from .utils import dtype_checker_df
from .utils import cleanup_df_values_for_given_dtype


class EOCubeAbstract():
    def __init__(self, df_layers, chunksize=2**5, wdir=None):
        self._df_layers = df_layers
        self._chunksize = chunksize
        # In this whole class we could directly get the windows from the
        # eobox.raster.windows_from_blocksize(blocksize_xy, width, height)
        # then we do not need the more expensive initialization of MultiRasterIO
        self._mrio = None
        self._initialize_mrio_if_attr_is_none()
        if wdir:
            self._wdir = Path(wdir)
        else:
            self._wdir = wdir

    @property
    def df_layers(self):
        return self._df_layers

    @property
    def chunksize(self):
        return self._chunksize

    @chunksize.setter
    def chunksize(self, value):
        self._mrio = self._mrio.windows_from_blocksize(value)
        self._chunksize = value

    @property
    def n_chunks(self):
        return len(self._mrio.windows)

    @property
    def wdir(self):
        return self._wdir

    @wdir.setter
    def wdir(self, value):
        self._wdir = value

    def _initialize_mrio_if_attr_is_none(self):
        if not self._mrio:
            self._mrio = MultiRasterIO(layer_files=self.df_layers["path"].values)
            self._mrio = self._mrio.windows_from_blocksize(self.chunksize)

    def get_chunk_path_from_layer_path(self, path_layer, ji, mkdir=True):
        path_layer = Path(path_layer)
        bdir_chunks = path_layer.parent / f"xchunks_cs{self.chunksize}" / path_layer.stem
        if mkdir:
            bdir_chunks.mkdir(exist_ok=True, parents=True)
        ndigits = len(str((self.n_chunks)))
        dst_path_chunk = bdir_chunks / (path_layer.stem + f"_ji-{ji:0{ndigits}d}.tif")
        return dst_path_chunk

    def get_df_ilocs(self, band, date):
        """Get positions of rows matching specific band(s) and date(s).

        The method supports three typical queries:

        * one band and one date (both given as strings)

        * one band and of several dates (band given as strings, date as list of strings)

        * several band and of one date (date given as strings, band as list of strings)

        Arguments:
            band {str or list} -- Band(s) for which to derive the iloc index.
            date {str or list} -- Date(s) for which to derive the iloc index.

        Returns:
            int or list -- Integer (if band and date are str) or list of iloc indices.
        """

        df = self.df_layers.copy()
        df["index"] = range(df.shape[0])

        idx_layers = []
        if isinstance(band, str) and isinstance(date, str):
            idx_layers = df[(df["date"] == date) &  (df["band"] == band)]["index"].values[0]
        if isinstance(band, list) and isinstance(date, str):
            for b in band:
                idx = df[(df["date"] == date) &  (df["band"] == b)]["index"].values[0]
                idx_layers.append(idx)
        elif isinstance(band, str) and isinstance(date, list):
            for d in date:
                idx = df[(df["band"] == band) &  (df["date"] == d)]["index"].values[0]
                idx_layers.append(idx)
        return idx_layers

class EOCube(EOCubeAbstract):
    def __init__(self, df_layers, chunksize=2**5, wdir=None):
        super().__init__(df_layers=df_layers, chunksize=chunksize, wdir=wdir)

    def get_chunk(self, ji):
        """Get a EOCubeChunk"""
        return EOCubeChunk.from_eocube(self, ji)

    def apply_and_write(self, fun, dst_paths, **kwargs):
        ji_process = []
        for ji in range(self.n_chunks):
            dst_paths_ji_exist = [self.get_chunk_path_from_layer_path(
                dst, ji, mkdir=False).exists() for dst in dst_paths]
            if not all(dst_paths_ji_exist):
                ji_process.append(ji)
        if len(ji_process) != self.n_chunks:
            print(f"{self.n_chunks - len(ji_process)} chunks already processed and skipped.")
        results = []
        for ji in tqdm(ji_process, total=len(ji_process)):
            eoc_chunk = self.get_chunk(ji)
            results.append(fun(eoc_chunk, dst_paths, **kwargs))
        for pth in dst_paths:
            if not Path(pth).exists():
                self.create_vrt_from_chunks(pth)

    def create_vrt_from_chunks(self, dst_path):
        chunk_paths = []
        for ji in range(self.n_chunks):
            pth = self.get_chunk_path_from_layer_path(dst_path, ji,
                                                      mkdir=True).absolute()
            if not pth.exists():
                raise FileNotFoundError(pth)
            else:
                chunk_paths.append(str(pth))
        buildvrt(chunk_paths, str(dst_path), relative=True)


class EOCubeChunk(EOCubeAbstract):
    def __init__(self, ji, df_layers, chunksize=2**5, wdir=None):
        super().__init__(df_layers=df_layers, chunksize=chunksize, wdir=wdir)
        self._ji = ji
        self._data = None # set with self.read_data()
        self._data_structure = None # can be ndarray, dataframe
        self._window = self._mrio.windows[self.ji]
        self._width = self._window.width
        self._height = self._window.height
        self._n_layers = self.df_layers.shape[0]
        self.result = None
        self._spatial_bounds = self._get_spatial_bounds()
        
    @property
    def ji(self):
        return self._ji

    @property
    def data(self):
        return self._data

    @property
    def chunksize(self):
        return self._chunksize

    @property
    def spatial_bounds(self):
        return self._spatial_bounds

    @chunksize.setter
    def chunksize(self, value):
        raise NotImplementedError("It is not allowed to set the EOCubeChunk chunksize (but of a EOCube object).")

    def _get_spatial_bounds(self):
        """Get the spatial bounds of the chunk.""" 
        # This should be a MultiRasterIO method
        with rasterio.open(self._mrio._get_template_for_given_resolution(self._mrio.dst_res, "path")) as src_layer:
            pass # later we need src_layer for src_layer.window_transform(win)
        win_transform = src_layer.window_transform(self._window)
        bounds = rasterio.windows.bounds(window=self._window,
                                         transform=win_transform,
                                         height=0, width=0)
        return bounds

    def read_data(self):
        self._data = self._mrio.get_arrays(self.ji)
        if self.data.shape[0] * self.data.shape[1] != self._width * self._height:
            raise Exception(f"X/Y dimension size of extracted window (={'/'.join(self.data.shape)}) different from expected shape (={self._width}/{self._height}).")
        self._update_data_structure()
        return self

    def _update_data_structure(self):
        self._data_structure = self.data.__class__.__name__

    def convert_data_to_dataframe(self):
        if self._data_structure != "ndarray":
            raise Exception(f"Data is not an ndarray but {self._data_structure}.")
        if "uname" not in self.df_layers.columns:
            raise Exception("You need a column named uname with unique names for all layers.")
        if not self.df_layers["uname"].is_unique:
            raise Exception("You need a column named uname with unique names for all layers.")

        self._data = pd.DataFrame(self._reshape_3to2d(self._data))
        self._update_data_structure()
        self._data.columns = self.df_layers["uname"]
        return self

    def convert_data_to_ndarray(self):
        """Converts the data from dataframe to ndarray format. Assumption: df-columns are ndarray-layers (3rd dim.)"""
        if self._data_structure != "DataFrame":
            raise Exception(f"Data is not a DataFrame but {self._data_structure}.")
        self._data = self._convert_to_ndarray(self._data)
        self._update_data_structure()
        return self

    def _convert_to_ndarray(self, data):
        """Converts data from dataframe to ndarray format. Assumption: df-columns are ndarray-layers (3rd dim.)"""
        if data.__class__.__name__ != "DataFrame":
            raise Exception(f"data is not a DataFrame but {data.__class__.__name__}.")
        shape_ndarray = (self._height, self._width, data.shape[1])
        data_ndarray = data.values.reshape(shape_ndarray)
        return data_ndarray

    def write_dataframe(self, result, dst_paths, nodata=None, compress='lzw'):
        """Write results (dataframe) to disc."""
        result = self._convert_to_ndarray(result)
        self.write_ndarray(result, dst_paths, nodata=nodata, compress=compress)

    def write_ndarray(self, result, dst_paths, nodata=None, compress='lzw'):
        """Write results (ndarray) to disc."""

        assert len(dst_paths) == result.shape[2]
        assert result.shape[0] == self._height
        assert result.shape[1] == self._width
        assert result.shape[2] == len(dst_paths)
        with rasterio.open(self._mrio._get_template_for_given_resolution(self._mrio.dst_res, "path")) as src_layer:
            pass # later we need src_layer for src_layer.window_transform(win)
        for i, pth in enumerate(dst_paths):
            dst_path_chunk = self.get_chunk_path_from_layer_path(pth, self.ji)

            result_layer_i = result[:, :, [i]]
            assert result_layer_i.shape[2] == 1
            kwargs = self._mrio._get_template_for_given_resolution(
                res=self._mrio.dst_res, return_="meta").copy()
            kwargs.update({"driver": "GTiff",
                           "compress": compress,
                           "nodata": nodata,
                           "height": self._height,
                           "width": self._width,
                           "dtype": result_layer_i.dtype,
                           "transform": src_layer.window_transform(self._window)})
            with rasterio.open(dst_path_chunk, "w", **kwargs) as dst:
                dst.write(result_layer_i[:, :, 0], 1)
    
    @staticmethod
    def robust_data_range(arr, robust=False, vmin=None, vmax=None):
        """Get a robust data range, i.e. 2nd and 98th percentile for vmin, vmax parameters."""
        # from the seaborn code 
        # https://github.com/mwaskom/seaborn/blob/3a3ec75befab52c02650c62772a90f8c23046038/seaborn/matrix.py#L201

        def _get_vmin_vmax(arr2d, vmin=None, vmax=None):
            if vmin is None:
                vmin = np.percentile(arr2d, 2) if robust else arr2d.min()
            if vmax is None:
                vmax = np.percentile(arr2d, 98) if robust else arr2d.max()
            return vmin, vmax

        if len(arr.shape) == 3 and vmin is None and vmax is None:
            vmin = []
            vmax = []
            for i in range(arr.shape[2]):
                arr_i = arr[:, :, i]
                vmin_i, vmax_i = _get_vmin_vmax(arr_i, vmin=None, vmax=None)
                vmin.append(vmin_i)
                vmax.append(vmax_i)
        else:
            vmin, vmax = _get_vmin_vmax(arr, vmin=vmin, vmax=vmax)
        return vmin, vmax

    def plot_raster(self,
                    idx_layer,
                    robust=False, vmin=None, vmax=None,
                    spatial_bounds=False,
                    figsize=None, ax=None):

        import matplotlib.pyplot as plt

        if self.data is None:
            self.read_data()
        arr = self.data[:, :, idx_layer]

        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        vmin, vmax = self.robust_data_range(arr,
                                            robust=robust,
                                            vmin=vmin, vmax=vmax)
        ax = ax.imshow(arr,
                       extent=self.spatial_bounds if spatial_bounds else None,
                       vmin=vmin, vmax=vmax)
        return ax

    def plot_raster_rgb(self,
                        idx_layers,
                        robust=False, vmin=None, vmax=None,
                        spatial_bounds=False,
                        figsize=None, ax=None):

        import matplotlib.pyplot as plt

        def _select_vmxx(vmxx, i):
            """If vmxx is a list, return i-th value of the list, else return vmxx."""
            if isinstance(vmxx, list):
                return vmxx[i]
            else:
                return vmxx

        if self.data is None:
            self.read_data()
        arr = self.data[:, :, idx_layers].astype(float)

        for i in range(arr.shape[2]):
            arr_i = arr[:, :, i]

            vmin_i, vmax_i = self.robust_data_range(arr_i, 
                                                    robust=robust, 
                                                    vmin=_select_vmxx(vmin, i),
                                                    vmax=_select_vmxx(vmax, i))
            arr_i[arr_i < vmin_i] = vmin_i
            arr_i[arr_i > vmax_i] = vmax_i
            arr[:, :, i] = (arr_i - vmin_i) / (vmax_i - vmin_i)

        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        ax = ax.imshow(arr, 
                       extent=self.spatial_bounds if spatial_bounds else None)

        return ax

    

    @staticmethod
    def _reshape_3to2d(arr_3d):
        new_shape = (arr_3d.shape[0] * arr_3d.shape[1], arr_3d.shape[2],)
        return arr_3d.reshape(new_shape)

    @staticmethod
    def from_eocube(eocube, ji):
        """Create a EOCubeChunk object from an EOCube object."""
        eocubewin = EOCubeChunk(ji, eocube.df_layers, eocube.chunksize, eocube.wdir)
        return eocubewin


class EOCubeSceneCollectionAbstract(EOCubeAbstract):
    def __init__(self,
                 df_layers,
                 chunksize,
                 variables,
                 qa,
                 qa_valid,
                 # timeless=None,
                 wdir=None):

        # validation and formatting
        n_per_sceneid = len(variables) + 1 # i.e. qa layer
        scenes_complete = df_layers.groupby("sceneid").apply(
            lambda x: x["band"].isin(variables + [qa]).sum() == n_per_sceneid)
        if not all(scenes_complete):
            scenes_incomplete = scenes_complete.index[~scenes_complete].values
            raise ValueError(f"Variable or qa layers missing in the following scenes: {scenes_incomplete}")
        df_layers = df_layers.sort_values(["date", "sceneid", "band"])

        EOCubeAbstract.__init__(self, df_layers=df_layers, chunksize=chunksize, wdir=wdir)
        self._variables = variables
        self._qa = qa
        self._qa_valid = qa_valid
        # self._timeless = timeless

    @property
    def variables(self):
        return self._variables

    @property
    def qa(self):
        return self._qa

    @property
    def qa_valid(self):
        return self._qa_valid

"""     @property
    def timeless(self):
        return self._timeless
 """

class EOCubeSceneCollection(EOCubeSceneCollectionAbstract, EOCube):

    def get_chunk(self, ji):
        """Get a EOCubeChunk"""
        return EOCubeSceneCollectionChunk(ji=ji,
                                          df_layers=self.df_layers,
                                          chunksize=self.chunksize,
                                          variables=self.variables,
                                          qa=self.qa,
                                          qa_valid=self.qa_valid,
                                          wdir=self.wdir)

    def apply_and_write_by_variable(self,
                                    fun,
                                    dst_paths,
                                    dtypes,
                                    compress,
                                    nodata,
                                    **kwargs):

        def check_variable_specific_args(arg, variables):
            if isinstance(arg, dict):
                if not set(arg.keys()) == set(self.variables):
                    raise ValueError("'dtypes'-, 'nodata'-dicts keys must match (self.)variables.")
                else:
                    return arg # TODO: better validate the args before
            else:
                # TODO: validate the args before
                return {var: arg for var in variables}

        dtypes = check_variable_specific_args(dtypes, self.variables)
        nodata = check_variable_specific_args(nodata, self.variables)
        compress = check_variable_specific_args(compress, self.variables)

        ji_process = {}
        for var in self.variables:
            ji_process[var] = []
            counter = 0
            for ji in range(self.n_chunks):
                dst_paths_ji_exist = [self.get_chunk_path_from_layer_path(
                    dst, ji, mkdir=False).exists() for dst in dst_paths[var]]
                if all(dst_paths_ji_exist):
                    counter += 1
                else:
                    ji_process[var].append(ji)
            if len(ji_process[var]) != self.n_chunks:
                print(f"{var}: {counter} / {self.n_chunks} chunks already processed and skipped.")

        # a sorted list of all chunks that we need to run somewhere
        ji_process_over_all_variables = []
        for key, process in ji_process.items():
            ji_process_over_all_variables += process
        ji_process_over_all_variables = np.unique(ji_process_over_all_variables)

        for ji in tqdm(ji_process_over_all_variables, total=len(ji_process)):
            eoc_chunk = self.get_chunk(ji)
            eoc_chunk = eoc_chunk.read_data_by_variable(mask=True)
            results = {}
            for var in self.variables:
                if ji in ji_process[var]:
                    results[var] = fun(eoc_chunk.data[var], **kwargs)
                    results[var] = cleanup_df_values_for_given_dtype(results[var],
                                                                     dtype=dtypes[var],
                                                                     lower_as=None,
                                                                     higher_as=None,
                                                                     nan_as=None)
                    eoc_chunk.write_dataframe(results[var],
                                              dst_paths[var],
                                              nodata[var],
                                              compress[var])
        for var in self.variables:
            for pth in dst_paths[var]:
                if not Path(pth).exists():
                    self.create_vrt_from_chunks(pth)

    def create_virtual_time_series(self,
                                   idx_virtual,
                                   dst_pattern, #"./xxx_uncontrolled/ls2008_vts4w/ls2008_vts4w_{date}_{var}.vrt"
                                   dtypes,
                                   compress="lzw",
                                   nodata=None,
                                   num_workers=1):
        dst_paths = {}
        for var in self.variables:
            dst_paths[var] = []
            for date in idx_virtual:
                dst_paths[var].append(dst_pattern.format(**{"var": var, "date": date.strftime("%Y-%m-%d")}))
        assert (len(idx_virtual) * len(self.variables)) == sum([len(dst_paths[var]) for var in self.variables])

        self.apply_and_write_by_variable(# mask=True,
                                         fun=create_virtual_time_series,
                                         dst_paths=dst_paths,
                                         dtypes=dtypes,
                                         compress=compress,
                                         nodata=nodata,
                                         idx_virtual=idx_virtual,
                                         num_workers=num_workers,
                                         verbosity=0)


class EOCubeSceneCollectionChunk(EOCubeSceneCollectionAbstract, EOCubeChunk):
    def __init__(self,
                 ji,
                 df_layers,
                 chunksize,
                 variables,
                 qa,
                 qa_valid,
                 # timeless=None,
                 wdir=None):
        EOCubeSceneCollectionAbstract.__init__(self,
                                               df_layers=df_layers,
                                               chunksize=chunksize,
                                               variables=variables,
                                               qa=qa,
                                               qa_valid=qa_valid,
                                               wdir=wdir)
        EOCubeChunk.__init__(self,
                             ji=ji,
                             df_layers=df_layers,
                             chunksize=chunksize,
                             wdir=wdir)

    def read_data_by_variable(self, mask=True):
        """Reads and masks (if desired) the data and converts it in one dataframe per variable."""
        def print_elapsed_time(start, last_stopped, prefix):
            # print(f"{prefix} - Elapsed time [s] since start / last stopped: \
            #     {(int(time.time() - start_time))} / {(int(time.time() - last_stopped))}")
            return time.time()
        start_time = time.time()
        last_stopped = time.time()
        last_stopped = print_elapsed_time(start_time, last_stopped, "Starting chunk function")

        verbose = False

        self.read_data()
        last_stopped = print_elapsed_time(start_time, last_stopped, "Data read")

        # 2.
        sc_chunk = self.convert_data_to_dataframe()
        last_stopped = print_elapsed_time(start_time, last_stopped, "Data converted to df")


        # 3.B.
        if mask:
            # 3.A.
            ilocs_qa = np.where((self.df_layers["band"] == self.qa).values)[0]
            df_qa = self.data.iloc[:, ilocs_qa]
            df_qa.columns = self.df_layers["date"].iloc[ilocs_qa]
            df_clearsky = df_qa.isin(self.qa_valid)
            last_stopped = print_elapsed_time(start_time, last_stopped, "Clearsky df created")

            return_bands = self.variables
        else:
            return_bands = self.variables + [self.qa]

        dfs_variables = {}
        for var in return_bands:
            if verbose:
                print("VARIABLE:", var)
            ilocs_var = np.where((self.df_layers["band"] == var).values)[0]
            df_var = self.data.iloc[:, ilocs_var]
            df_var.columns = self.df_layers["date"].iloc[ilocs_var]
            if mask:
                df_var = df_var.where(df_clearsky, other=np.nan)
            dfs_variables[var] = df_var
        last_stopped = print_elapsed_time(start_time, last_stopped, "Clearsky df created")
        self._data = dfs_variables
        return self


def create_virtual_time_series(df_var, idx_virtual, num_workers=1, verbosity=0):

    def _create_virtual_time_series_core(df, idx_virtual, verbosity=0):

        if verbosity:
            print("Shape of input dataframe:", df.shape)

        transpose_before_return = False
        if isinstance(df.columns, pd.DatetimeIndex):
            transpose_before_return = True
            df = df.transpose()

        # define the virtual points in the time series index
        # idx_virtual = df.resample(rule).asfreq().index
        if not df.index.is_unique:
            if verbosity:
                print(f"Aggregating (max) data with > observations per day: {df.index[df.index.duplicated()]}")
            df = df.groupby(level=0).max()
            assert df.index.is_unique

        if verbosity:
            print("Length of virtual time series:", len(idx_virtual))
        # add the existing time series points to the virtual time series index
        idx_virtual_and_data = idx_virtual.append(df.index).unique()
        idx_virtual_and_data = idx_virtual_and_data.sort_values()
        if verbosity:
            print("Length of virtual and data time series:", len(idx_virtual_and_data))
        # extend the time series data such that it contains all existing and virtual time series points
        df = df.reindex(index=idx_virtual_and_data)
        # interpolate between dates and forward/backward fill edges with closest values
        df = df.interpolate(method='time')
        df = df.bfill()
        df = df.ffill()
        df = df.loc[idx_virtual]

        if transpose_before_return:
            df = df.transpose()
        if verbosity:
            print("Shape of output dataframe:", df.shape)
        return df

    if (num_workers > 1) or (num_workers == -1):
        import dask.dataframe as dd
        df_var = dd.from_pandas(df_var, npartitions=num_workers)
        df_result = df_var.map_partitions(_create_virtual_time_series_core,
                                          idx_virtual=idx_virtual,
                                          verbosity=verbosity)
        df_result = df_result.compute(scheduler='processes', num_workers=num_workers)
    else:
        df_result = _create_virtual_time_series_core(df_var,
                                                     idx_virtual,
                                                     verbosity=verbosity)
    return df_result



"""
eobox.vector subpackage

Subpackage which mainly builds upon shapely, fiona and geopandas (and thus ogr) functionality.
It provides tools for 

* clean convertion of polygons to lines.


For more information on the package content, visit [readthedocs](https://eo-box.readthedocs.raster/en/latest/eobox.vector.html).

"""

from . import conversion

from .conversion import calc_distance_to_border
from .conversion import convert_polygons_to_lines



import os
from setuptools import setup, find_packages
import subprocess


def get_long_description():
    this_directory = os.path.abspath(os.path.dirname(__file__))

    with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()

    return long_description

def parse_requirements(file):
    return sorted(set(
        line.partition('#')[0].strip()
        for line in open(os.path.join(os.path.dirname(__file__), file))
    ) - set(''))


def get_version():
    for line in open(os.path.join(os.path.dirname(__file__), 'eobox', '__init__.py')):
        if line.find("__version__") >= 0:
            version = line.split("=")[1].strip()
            version = version.strip('"').strip("'")
    return version

setup(name='eobox',
      python_requires='>=3.5,<3.7',
      version=get_version(),
      description='A toolbox for processing earth observation data with Python.',
      long_description=get_long_description(),
      long_description_content_type='text/markdown',
      url='https://github.com/benmack/eo-box',
      author='Benjamin Mack',
      author_email='ben8mack@gmail.com',
      license='MIT',
      packages=find_packages(),
      install_requires=parse_requirements("requirements.txt"),
      include_package_data=True,
      package_data={'eobox/sampledata/data/s2l1c': [
          's2l1c/**/*jp2', 
          's2l1c/s2l1c_ref.gpkg',
          's2l1c/gis_osm_landuse-water_a_free_1_area-10000-to-500000.gpkg'
          ],
          'eobox/sampledata/data/lsts': ['lsts/**/*tif']
          },
      zip_safe=False)



src_file = "/media/ben/AO/phd/01_eodata/preprocessed/s1ms_ms_10m/33UUU_2017/VH/VH_02K0001_20170102T165141_S1B_db.tif"
dst_file = "/home/ben/Desktop/33UUU_2017/VH/VH_02K0001_20170102T165141_S1B_db.tif"
template_file = "/media/ben/AO/phd/01_eodata/readytouse/s2__33UUU_2017__2w__ndvi__vts/33UUU_2017/s2_uuu33_vts2w_2017-01-01_ndvi.tif"


from eobox.raster import gdalutils

gdalutils.reproject_on_template_raster(
    src_file, dst_file, template_file, resampling="near", compress="deflate"
)


# from pathlib import Path
# import pandas as pd
# import rasterio as rio
# import numpy as np
# from glob import glob
# import os
# from osgeo import gdal
# from tqdm import tqdm
# import subprocess

# from eobox.raster import cube
# from eobox.raster import gdalutils

# from eobox.raster.utils import cleanup_df_values_for_given_dtype
# from eobox.raster.utils import dtype_checker_df

# from agrotools import configs

# prjconf = configs.ProjectConfigParser(config_file=None)

# src_ms = prjconf.get_path('EOdata', 'preprocessed')
# dst_ms = prjconf.get_path('EOdata', 'readytouse')

# src_ms_vv =  Path(src_ms / 's1ms_ms_10m/33UUU_2017/VV')
# src_ms_vh =  Path(src_ms / 's1ms_ms_10m/33UUU_2017/VH')

# dst_ms_vv = Path(dst_ms / 's1ms__33UUU_2017__2w__VV__median')
# dst_ms_vh = Path(dst_ms / 's1ms__33UUU_2017__2w__VH__median')

# assert src_ms_vv.exists() & src_ms_vh.exists()

# list_vv = list(src_ms_vv.glob("**/VV_08K0001_*S1[A,B]_db.tif"))
# list_vh = list(src_ms_vh.glob("**/VH_02K0001_*S1[A,B]_db.tif"))
# print("Number of scenes for processing: ", (len(list_vv) + len(list_vh)))
# print(list_vv[:3])
# print(list_vh[:3])
# list_vvvh = (list_vh + list_vv)
# print(len(list_vvvh))

# dct ={}

# for scene in list_vvvh:
#     splitted_stems = scene.stem.split('_')
#     dct[scene] = [scene, scene.stem, splitted_stems[0], splitted_stems[2], splitted_stems[3]]
# df_vvvh = pd.DataFrame.from_dict(dct).T.reset_index(drop=True).rename(columns={0:'path', 1:'uname', 2:'polarization', 3:'date', 4:'sensor'})
# df_vvvh['date'] = pd.to_datetime(df_vvvh['date'].str[:-7])
# df_vvvh['week_number'] =  df_vvvh['date'].dt.week
# df_vvvh['path'] =  df_vvvh['path'].astype(np.str)
# df_vvvh = df_vvvh.sort_values('date')

# print(df_vvvh.shape)

# biweekly_index = pd.date_range(start='2017-01-01', end='2017-12-31', freq='2W')
# df_vvvh['biweek_number'] = 0

# for i in range(0, 27, 1):
#     if i == 26:
#         break
#     else:
#         df_vvvh.loc[(df_vvvh['date'] > biweekly_index[i]) & (df_vvvh['date'] <= biweekly_index[i+1]), ['biweek_number']] = i+1

# assert df_vvvh.week_number.values.max() == 52
# assert df_vvvh.biweek_number.values.max() == 26

# df_vv = df_vvvh[df_vvvh['polarization']=='VV']
# df_vh = df_vvvh[df_vvvh['polarization']=='VH']

# assert df_vv.shape == df_vh.shape

# biweekly_vv = list(df_vv.groupby('biweek_number')['path'].apply(list))
# biweekly_vh = list(df_vh.groupby('biweek_number')['path'].apply(list))

# df_tmp = df_vh[df_vvvh['biweek_number']==1]

# #import rasterio
# #for i in range(df_tmp.shape[0]):
# #    with rasterio.open(df_tmp.path.iloc[0]) as src:
# #        print(src.meta["width"], src.meta["height"])

# # eoc._mrio.windows

# df_l = df_tmp.reset_index(drop=True)

# eoc = cube.EOCube(df_layers=df_l.loc[[5,6], :], chunksize=2**11)
# dst_paths = ["/home/ben/Desktop/a_median_dir/a_median.vrt"]

# ji = 1

# eoc_chunk = eoc.get_chunk(ji=ji)
# eoc_chunk = eoc_chunk.read_data()
# eoc_chunk.data.shape[0] * eoc_chunk.data.shape[1] == eoc_chunk._width * eoc_chunk._height
